//////////////////////////////////////////////////////////////////////////////
//
//      Copyright  1999 Zentropic Computing, All rights reserved
//
// Authors:             Steve Papacharalambous
// Original date:       Thu 15 Jul 1999
//
// pthreads interface for Real Time Linux.
//
//
// The contents of this file are subject to the Mozilla Public
// License Version 1.0 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////

static char id_rtai_pthread_c[] __attribute__ ((unused)) = "@(#)$Id: rtai_pthread.c,v 1.2 1999/09/03 09:54:28 stevepapa Exp $";


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/time.h>

#include <rtai_pthread.h>
#include <rtai_pthread_int.h>


#define RUNNABLE_ON_CPUS 3
#define RUN_ON_CPUS (smp_num_cpus > 1 ? RUNNABLE_ON_CPUS : 1)

#define TIMER_TO_CPU 3             // < 0 || > 1 to maintain a symmetrically processed timer

#define RT_STACK 3000

#define DEFAULT_TICK_PERIOD 20000

typedef void (* RT_THREAD)(int);

// ----------------------------------------------------------------------------

static char pthread_version[] = "0.1";

static unsigned int num_pthreads = 0;
static struct _pthread_descr_struct rt_pthread_descr[POSIX_THREADS_MAX] = {{0}};
static RTIME tick_period;


// ----------------------------------------------------------------------------

// Kludge section.
// Replicates code sections that are not easily accessible from RT land.
// Hmm..  there must be a better way....

#define PTHREAD_PRIORITY_FACTOR 99

// Linux has no ENOTSUP error code.
#define ENOTSUP EOPNOTSUPP


int get_max_priority(int policy) {

  int ret = -EINVAL;

  switch (policy) {
    case SCHED_FIFO:
    case SCHED_RR:
      ret = 99;
      break;
    case SCHED_OTHER:
      ret = 0;
      break;
    default:
      break;
  }

  return(ret);

}  // End function - get_max_priority.


int get_min_priority(int policy) {

  int ret = -EINVAL;

  switch (policy) {
    case SCHED_FIFO:
    case SCHED_RR:
      ret = 1;
      break;
    case SCHED_OTHER:
      ret = 0;
      break;
    default:
      break;
  }

  return ret;

}  // End function - get_min_priority.


// ----------------------------------------------------------------------------


// Thread creation/control section.


int pthread_create(pthread_t *thread, pthread_attr_t *attr,
                    void *(*start_routine) (void *), void *arg) {

  int i, j, r_c;

  if( num_pthreads >= POSIX_THREADS_MAX ) {
    return(EAGAIN);

  }  // End if - Cannot create any more pthreads.
// Find a free pthread descriptor structure.
// Note: A value of 0 in the p_tid field indicates a free pthread descriptor.
  for(j = 0; j < POSIX_THREADS_MAX; j++) {
    if(rt_pthread_descr[j].p_tid == 0) {
// Initialise the thread descriptor.
      rt_pthread_descr[j].p_nextwaiting = NULL;
      rt_pthread_descr[j].p_tid = j + 1;
      rt_pthread_descr[j].p_priority = 0; // Default to lowest pthread priority.
      rt_pthread_descr[j].p_policy = SCHED_OTHER;
      rt_pthread_descr[j].p_spinlock = NULL;
      rt_pthread_descr[j].p_signal = 0;
      rt_pthread_descr[j].p_signal_jmp = NULL;
      rt_pthread_descr[j].p_cancel_jmp = NULL;
      rt_pthread_descr[j].p_terminated = 0;
      rt_pthread_descr[j].p_detached = attr == NULL ? 0 : attr->detachstate;
      rt_pthread_descr[j].p_exited = 0;
      rt_pthread_descr[j].p_retval = NULL;
      rt_pthread_descr[j].p_joining = NULL;
      rt_pthread_descr[j].p_cleanup = NULL;
      rt_pthread_descr[j].p_cancelstate = PTHREAD_CANCEL_ENABLE;
      rt_pthread_descr[j].p_canceltype = PTHREAD_CANCEL_DEFERRED;
      rt_pthread_descr[j].p_canceled = 0;
      rt_pthread_descr[j].p_errno = 0;
      rt_pthread_descr[j].p_h_errno = 0;
      for(i = 0; i < PTHREAD_KEY_1STLEVEL_SIZE; i++) {
        rt_pthread_descr[j].p_specific[i] = NULL;
      }
      rt_pthread_descr[j].p_start_args.schedpolicy = SCHED_OTHER;

// Initialise the scheduling parameters for the new thread.
      if(attr != NULL && attr->schedpolicy != SCHED_OTHER) {
        switch(attr->inheritsched) {
          case PTHREAD_EXPLICIT_SCHED:
            rt_pthread_descr[j].p_start_args.schedpolicy = attr->schedpolicy;
            rt_pthread_descr[j].p_start_args.schedparam = attr->schedparam;
          break;
          case PTHREAD_INHERIT_SCHED:
// TODO
//These must be uncommented when the new RT calls have been implemented,
// also father_pid has to be determined.
// stevep 19 July 1999
//        rt_pthread_descr[j].schedpolicy = sched_getscheduler(father_pid);
//        sched_getparam(father_pid, &rt_pthread_descr[j].schedparam);
          break;
        }
        rt_pthread_descr[j].p_priority =
          rt_pthread_descr[j].p_start_args.schedparam.sched_priority;
        rt_pthread_descr[j].p_policy =
          rt_pthread_descr[j].p_start_args.schedpolicy;
      }
      rt_pthread_descr[j].p_start_args.start_routine = start_routine;
      rt_pthread_descr[j].p_start_args.arg = arg;
//TODO
// Not currently set up, but needed for full implementation?
//  rt_pthread_descr[j].p_start_args.mask = *mask;

// Initialise the priority in the RTAI task structure.
// The priorities in RTAI are the other way up to the POSIX ones, sigh...
      if(rt_pthread_descr[j].p_priority > PTHREAD_PRIORITY_FACTOR) {
        rt_pthread_descr[j].p_priority = PTHREAD_PRIORITY_FACTOR;
      }
      rt_pthread_descr[j].rtask_struct.priority =
                   PTHREAD_PRIORITY_FACTOR - rt_pthread_descr[j].p_priority;

// Save the pointer to this pthread descriptor in the RT scheduler structure.
      rt_pthread_descr[j].rtask_struct.system_data_ptr = (void *)&rt_pthread_descr[j];

// Create the new thread.
      r_c = rt_task_init(&rt_pthread_descr[j].rtask_struct,
                         (RT_THREAD)start_routine, (int)arg, RT_STACK,
                         rt_pthread_descr[j].rtask_struct.priority,
                         TASK_FPU_ENABLE, NULL);

      if( r_c == 0 ) {

        rt_pthread_descr[j].p_tid = j + 1;
        rt_set_runnable_on_cpus(&rt_pthread_descr[j].rtask_struct,
                                RUN_ON_CPUS);

        if( rt_task_resume(&rt_pthread_descr[j].rtask_struct) < 0 ) {
// Uurgh -- should be: -EINVAL, but only error return specified for
// pthread_create is EAGAIN.
          return(EAGAIN);
        }  // End if - Error from rt_task_resume

// Return the index (+1) into the attribute array as the new thread id.
        *thread = (pthread_t)j + 1;
        return(0);
      } else {

        return(EAGAIN);
      }  // End if/else - process return code from rt_task_init

    }  // End if - Initialise pthread descriptor.

  }  // End for loop - find a free pthread descriptor.

  return(EAGAIN);

}  // End function - pthread_create


void pthread_exit(void *retval) {

  pthread_descr thread_ptr;
  RT_TASK *rt_task_ptr;

  rt_task_ptr = rt_whoami();
  thread_ptr = (pthread_descr)rt_task_ptr->system_data_ptr;
  thread_ptr->p_tid = 0;
  rt_task_delete(rt_task_ptr);

} // End function - pthread_exit


int sched_yield(void) {

  rt_task_yield();
  return(0);

}  // End function - pthread_yield.


pthread_t pthread_self(void) {

  pthread_descr thread_ptr;
  RT_TASK *rt_task_ptr;

  rt_task_ptr = rt_whoami();
  thread_ptr = (pthread_descr)rt_task_ptr->system_data_ptr;
  return(thread_ptr->p_tid);

} // End function - pthread_self


// ----------------------------------------------------------------------------

// Thread attribute section.

int pthread_attr_init(pthread_attr_t *attr) {

  attr->detachstate = PTHREAD_CREATE_JOINABLE;
  attr->schedpolicy = SCHED_OTHER;
  attr->schedparam.sched_priority = 0;
  attr->inheritsched = PTHREAD_EXPLICIT_SCHED;
  attr->scope = PTHREAD_SCOPE_SYSTEM;
  return(0);

}  // End function - pthread_attr_init


int pthread_attr_destroy(pthread_attr_t *attr) {

  return(0);

}  // End function - pthread_attr_destroy


int pthread_attr_setdetachstate(pthread_attr_t *attr, int detachstate) {

  if (detachstate < PTHREAD_CREATE_JOINABLE ||
      detachstate > PTHREAD_CREATE_DETACHED)
    return(EINVAL);
  attr->detachstate = detachstate;
  return(0);

}  // End function - pthread_attr_setdetachstate


int pthread_attr_getdetachstate(const pthread_attr_t *attr, int *detachstate) {

  *detachstate = attr->detachstate;
  return(0);

}  // End function - pthread_attr_getdetachstate


int pthread_attr_setschedparam(pthread_attr_t *attr,
			       const struct sched_param *param) {

  int max_prio = get_max_priority(attr->schedpolicy);
  int min_prio = get_min_priority(attr->schedpolicy);

  if(param->sched_priority < min_prio || param->sched_priority > max_prio) {
    return(EINVAL);
  }
  attr->schedparam = *param;
  return(0);

}  // End function - pthread_attr_setschedparam


int pthread_attr_getschedparam(const pthread_attr_t *attr,
			       struct sched_param *param) {

  *param = attr->schedparam;
  return(0);

}  // End function - pthread_attr_getschedparam


int pthread_attr_setschedpolicy(pthread_attr_t *attr, int policy) {

  if(policy != SCHED_OTHER && policy != SCHED_FIFO && policy != SCHED_RR) {
    return(EINVAL);
  }
  attr->schedpolicy = policy;
  return(0);

}  // End function - pthread_attr_setschedpolicy


int pthread_attr_getschedpolicy(const pthread_attr_t *attr, int *policy) {

  *policy = attr->schedpolicy;
  return(0);

}  // End function - pthread_attr_getschedpolicy


int pthread_attr_setinheritsched(pthread_attr_t *attr, int inherit) {

  if(inherit != PTHREAD_INHERIT_SCHED && inherit != PTHREAD_EXPLICIT_SCHED) {
    return(EINVAL);
  }
  attr->inheritsched = inherit;
  return(0);

}  // End function - pthread_attr_setinheritsched


int pthread_attr_getinheritsched(const pthread_attr_t *attr, int *inherit) {

  *inherit = attr->inheritsched;
  return(0);

}  // End function - pthread_attr_getinheritsched


int pthread_attr_setscope(pthread_attr_t *attr, int scope) {

  switch (scope) {
  case PTHREAD_SCOPE_SYSTEM:
    attr->scope = scope;
    return(0);
  case PTHREAD_SCOPE_PROCESS:
    return(ENOTSUP);
  default:
    return(EINVAL);
  }

}  // End function - pthread_attr_setscope


int pthread_attr_getscope(const pthread_attr_t *attr, int *scope) {

  *scope = attr->scope;
  return(0);

}  // End function - pthread_attr_getscope


// ----------------------------------------------------------------------------

// Thread scheduling section.


int pthread_setschedparam(pthread_t thread, int policy,
                          const struct sched_param *param) {

  int max_prio;
  int min_prio;
  unsigned long flags;
  pthread_descr pthread_handle;


// Get the thread descriptor and verify that it is a valid thread.
  pthread_handle = &rt_pthread_descr[thread - 1];
  if(pthread_handle->p_tid == 0) {
    return(ESRCH);
  }

  max_prio = get_max_priority(policy);
  min_prio = get_min_priority(policy);

  if(min_prio < 0 || max_prio < 0) {
    return(EINVAL);
  }

  if(param->sched_priority < min_prio || param->sched_priority > max_prio) {
    return(EINVAL);
  }

// Set the new scheduling parameters for the thread.
  r_save_flags(flags);
  r_cli();
  pthread_handle->p_priority = param->sched_priority;
  pthread_handle->p_policy = policy;
  pthread_handle->rtask_struct.priority = PTHREAD_PRIORITY_FACTOR - pthread_handle->p_priority;
  pthread_handle->rtask_struct.base_priority = PTHREAD_PRIORITY_FACTOR - pthread_handle->p_priority;
  r_restore_flags(flags);
  return(0);

} // End function - pthread_setschedparam


int pthread_getschedparam(pthread_t thread, int *policy, struct sched_param *param) {

  pthread_descr pthread_handle;


// Get the thread descriptor and verify that it is a valid thread.
  pthread_handle = &rt_pthread_descr[thread - 1];
  if(pthread_handle->p_tid == 0) {
    return(ESRCH);
  }

// Get the new scheduling parameters for the thread.
  *policy = pthread_handle->p_policy;
  param->sched_priority = pthread_handle->p_priority;
  return(0);

} // End function - pthread_getschedparam



// ----------------------------------------------------------------------------

// clock/timer section.

void clock_gettime(int clockid, struct timespec *current_time) {


  if(clockid != CLOCK_REALTIME) {
    current_time->tv_sec = 0;
    current_time->tv_nsec = 0;
  }

  ts_from_ns(count2nano(rt_get_time()), current_time);

}  // End function - clock_gettime


// ----------------------------------------------------------------------------

//  RT Task Utility Functions.


// Insert RT task into mutex queue in priority order.
void priority_enqueue_task(pthread_mutex_t *mutex) {

  QUEUE *rt_q;
  RT_TASK *current_task;

  rt_q = &(mutex->m_semaphore.queue);

// Check that the queues have been properly initialised.  Hackety hack... due to not
// being able to initialise the queues if the static initialiser macro is used...
  if(rt_q->prev == NULL || rt_q->next == NULL) {
    rt_q->prev = &(mutex->m_semaphore.queue);
    rt_q->next = &(mutex->m_semaphore.queue);

  }

  current_task = rt_whoami();

  while((rt_q = rt_q->next) != &(mutex->m_semaphore.queue) &&
          (rt_q->task)->priority <= current_task->priority);

  current_task->queue.prev = rt_q->prev;
  current_task->queue.next = rt_q;
  (rt_q->prev)->next = &(current_task->queue);
  rt_q->prev = &(current_task->queue);

}  // End function - priority_enqueue_task



// Insert RT task into conditional variable queue.
void cond_enqueue_task(pthread_cond_t *cond) {

  QUEUE *rt_q;
  RT_TASK *current_task;

  rt_q = &(cond->c_waiting.queue);

// Check that the queues have been properly initialised.  Hackety hack... due to not
// being able to initialise the queues if the static initialiser macro is used...
  if(rt_q->prev == NULL || rt_q->next == NULL) {
    rt_q->prev = &(cond->c_waiting.queue);
    rt_q->next = &(cond->c_waiting.queue);

  }

  current_task = rt_whoami();
  current_task->queue.prev = rt_q->prev;
  current_task->queue.next = rt_q;
  (rt_q->prev)->next = &(current_task->queue);
  rt_q->prev = &(current_task->queue);

}  // End function - cond_enqueue_task


// Propagate priority to task owning mutex.
void mutex_inherit_prio(pthread_mutex_t *mutex) {

  RT_TASK *current_task;

  current_task = rt_whoami();

  if(mutex->m_owner->priority > current_task->priority) {
    mutex->m_owner->priority = current_task->priority;
  }

}  // End function - mutex_inherit_prio



// Dequeue next task from the mutex queue.
void dequeue_task(pthread_mutex_t *mutex) {

  RT_TASK *next_task;

  next_task = (mutex->m_semaphore.queue.next)->task;
  (next_task->queue.prev)->next = next_task->queue.next;
  (next_task->queue.next)->prev = next_task->queue.prev;

}  // End function - dequeue_task


// Debug Routine - Display the tasks queued on a mutex queue.
void z_print_mutex_q(pthread_mutex_t *mutex) {

  QUEUE *rt_q;
  pthread_descr descr_ptr;
  
  rt_q = &(mutex->m_semaphore.queue);

  printk("--<z_print_mutex_q - RT threads on mutex queue>--\n");
  while((rt_q = rt_q->next) != &(mutex->m_semaphore.queue)) {
    descr_ptr = (pthread_descr)rt_q->task->system_data_ptr;
    printk("Task id: %ld, Priority: %d, State: %d \n", descr_ptr->p_tid, rt_q->task->priority, rt_q->task->state);

  }

  printk("--<End of thread list>--\n\n");

}  // End function - z_print_mutex_q



// ----------------------------------------------------------------------------

// Mutex section.

int pthread_mutex_init(pthread_mutex_t *mutex,
                       const pthread_mutexattr_t *mutex_attr) {

  QUEUE *rt_q;

  mutex->m_owner = 0;
  mutex->m_kind =
    mutex_attr == NULL ? PTHREAD_MUTEX_FAST_NP : mutex_attr->mutexkind;
  mutex->m_semaphore.magic = RT_SEM_MAGIC;
  mutex->m_semaphore.count = 1;
  rt_q = &(mutex->m_semaphore.queue);
  rt_q->prev = &(mutex->m_semaphore.queue);
  rt_q->next = &(mutex->m_semaphore.queue);
  rt_q->task = 0;
  return(0);

}  // End function - pthread_mutex_init


int pthread_mutex_destroy(pthread_mutex_t *mutex) {

  int return_code;
  unsigned long flags;
  QUEUE *rt_q;

  if(mutex->m_semaphore.magic != RT_SEM_MAGIC) {
    return(EINVAL);
  }
  r_save_flags(flags);
  r_cli();
  if(mutex->m_semaphore.count < 1) {
    return_code = EBUSY;
  } else {
    mutex->m_semaphore.magic = 0;
    rt_q = &(mutex->m_semaphore.queue);
    while ((rt_q = rt_q->next) != &(mutex->m_semaphore.queue)) {
      (rt_q->task)->state &= ~(SEMAPHORE | DELAYED);
    }

    return_code = 0;

  }  // End else - mutex is free, clear its queue.

  r_restore_flags(flags);
  return(return_code);

}  // End function - pthread_mutex_destroy



int pthread_mutexattr_init(pthread_mutexattr_t *attr) {

  attr->mutexkind = PTHREAD_MUTEX_FAST_NP;
  return(0);

}  // End function - pthread_mutexattr_init



int pthread_mutexattr_destroy(pthread_mutexattr_t *attr) {

  return(0);

}  // End function - pthread_mutex_attr_destroy


int pthread_mutexattr_setkind_np(pthread_mutexattr_t *attr, int kind) {

  if (kind != PTHREAD_MUTEX_FAST_NP
      && kind != PTHREAD_MUTEX_RECURSIVE_NP
      && kind != PTHREAD_MUTEX_ERRORCHECK_NP)
    return(EINVAL);
  attr->mutexkind = kind;
  return(0);

}  // End function - pthread_mutexattr_setkind_np


int pthread_mutexattr_getkind_np(const pthread_mutexattr_t *attr, int *kind) {

  *kind = attr->mutexkind;
  return(0);

}  // End function - pthread_mutexattr_getkind_np


int pthread_mutex_trylock(pthread_mutex_t *mutex) {

  int return_code;
  unsigned long flags;
  RT_TASK *self;

  if(mutex->m_semaphore.magic != RT_SEM_MAGIC) {
    return(EINVAL);
  }
  self = rt_whoami();
  r_save_flags(flags);
  r_cli();
  switch(mutex->m_kind) {
  case PTHREAD_MUTEX_FAST_NP:
    if(mutex->m_semaphore.count > 0) {
      mutex->m_semaphore.count = 0;
      mutex->m_owner = self;
      return_code = 0;
    } else {
      return_code = EBUSY;
    }
    break;
  case PTHREAD_MUTEX_RECURSIVE_NP:
    if (mutex->m_semaphore.count == 0 || mutex->m_owner == self) {
      mutex->m_semaphore.count--;
      mutex->m_owner = self;
      return_code = 0;
    } else {
      return_code = EBUSY;
    }
    break;
  case PTHREAD_MUTEX_ERRORCHECK_NP:
    if(mutex->m_semaphore.count > 0) {
      mutex->m_semaphore.count = 0;
      mutex->m_owner = self;
      return_code = 0;
    } else {
      return_code = EBUSY;
    }
    break;
  default:
    return_code = EINVAL;
  }

  r_restore_flags(flags);
  return(return_code);

}  // End function - pthread_mutex_trylock



int pthread_mutex_lock(pthread_mutex_t *mutex) {

  int return_code;
  unsigned long flags;
  RT_TASK *current_task;

  if(mutex->m_semaphore.magic != RT_SEM_MAGIC) {
    return(EINVAL);
  }
  current_task = rt_whoami();
  r_save_flags(flags);
  r_cli();
  switch(mutex->m_kind) {
  case PTHREAD_MUTEX_FAST_NP:
    if(mutex->m_semaphore.count > 0) {
      mutex->m_semaphore.count = 0;        // Resource is free, grab it.
      mutex->m_owner = current_task;
      return_code = 0;
    } else {
      current_task->state |= SEMAPHORE;
      current_task->blocked_on.sem = &(mutex->m_semaphore);
      priority_enqueue_task(mutex);
      mutex_inherit_prio(mutex);
      rt_schedule();
      if(current_task->blocked_on.sem ||
         mutex->m_semaphore.magic != RT_SEM_MAGIC) {
          return_code = EINVAL;
      } else {
        mutex->m_semaphore.count = 0;        // Resource is free now, grab it.
        mutex->m_owner = current_task;
        return_code = 0;
      }
    }
    break;
  case PTHREAD_MUTEX_RECURSIVE_NP:
    if(mutex->m_semaphore.magic != RT_SEM_MAGIC) {
      return(EINVAL);
    }
    if (mutex->m_semaphore.count == 0 || mutex->m_owner == current_task) {
      mutex->m_semaphore.count--;
      mutex->m_owner = current_task;
      return_code = 0;
    } else {
      current_task->state |= SEMAPHORE;
      current_task->blocked_on.sem = &(mutex->m_semaphore);
      priority_enqueue_task(mutex);
      mutex_inherit_prio(mutex);
      rt_schedule();
      if(current_task->blocked_on.sem ||
         mutex->m_semaphore.magic != RT_SEM_MAGIC) {
          return_code = EINVAL;
      } else {
        mutex->m_semaphore.count = 0;        // Resource is free now, grab it.
        mutex->m_owner = current_task;
        return_code = 0;
      }
    }
    break;
  case PTHREAD_MUTEX_ERRORCHECK_NP:
    if(mutex->m_semaphore.count > 0) {
      mutex->m_semaphore.count = 0;
      mutex->m_owner = current_task;
      return_code = 0;
    } else {
      current_task->state |= SEMAPHORE;
      current_task->blocked_on.sem = &(mutex->m_semaphore);
      priority_enqueue_task(mutex);
      mutex_inherit_prio(mutex);
      rt_schedule();
      if(current_task->blocked_on.sem ||
         mutex->m_semaphore.magic != RT_SEM_MAGIC) {
          return_code = EINVAL;
      } else {
        mutex->m_semaphore.count = 0;        // Resource is free now, grab it.
        mutex->m_owner = current_task;
        return_code = 0;
      }
    }
    break;
  default:
    return_code = EINVAL;
  }

  r_restore_flags(flags);
  return(return_code);

}  // End function - pthread_mutex_lock


int pthread_mutex_unlock(pthread_mutex_t *mutex) {

  int return_code = EINVAL;
  unsigned long flags;
  RT_TASK *current_task;

  current_task = rt_whoami();
  if(mutex->m_semaphore.magic != RT_SEM_MAGIC ||
                              mutex->m_owner != current_task) {
    return(EINVAL);
  }
  r_save_flags(flags);
  r_cli();
  switch(mutex->m_kind) {
  case PTHREAD_MUTEX_FAST_NP:
    mutex->m_semaphore.count = 1;
    mutex->m_owner = 0;
    current_task->priority = current_task->base_priority;

// Dequeue the next task in the queue.
    if(mutex->m_semaphore.queue.next != &(mutex->m_semaphore.queue) &&
       mutex->m_semaphore.queue.next != NULL && mutex->m_semaphore.queue.prev != NULL) {
      mutex->m_semaphore.queue.next->task->blocked_on.sem = 0;
      mutex->m_semaphore.queue.next->task->state &= ~(SEMAPHORE);
      dequeue_task(mutex);
    }
    rt_schedule();
    return_code = 0;
    break;
  case PTHREAD_MUTEX_RECURSIVE_NP:
    mutex->m_semaphore.count++;
    if(mutex->m_semaphore.count <= 0) {
      return_code = 0;
    } else {
      mutex->m_semaphore.count = 1;
      mutex->m_owner = 0;
      current_task->priority = current_task->base_priority;

// Dequeue the next task in the queue.
      if(mutex->m_semaphore.queue.next != &(mutex->m_semaphore.queue) &&
         mutex->m_semaphore.queue.next != NULL && mutex->m_semaphore.queue.prev != NULL) {
        mutex->m_semaphore.queue.next->task->blocked_on.sem = 0;
        mutex->m_semaphore.queue.next->task->state &= ~(SEMAPHORE);
        dequeue_task(mutex);
      }
      rt_schedule();
      return_code = 0;
    }
    break;
  case PTHREAD_MUTEX_ERRORCHECK_NP:
    if(mutex->m_semaphore.count == 1 || mutex->m_owner != current_task) {
      return_code = EPERM;
    } else {
      mutex->m_semaphore.count = 1;
      mutex->m_owner = 0;
      current_task->priority = current_task->base_priority;

// Dequeue the next task in the queue.
      if(mutex->m_semaphore.queue.next != &(mutex->m_semaphore.queue) &&
         mutex->m_semaphore.queue.next != NULL && mutex->m_semaphore.queue.prev != NULL) {
        mutex->m_semaphore.queue.next->task->blocked_on.sem = 0;
        mutex->m_semaphore.queue.next->task->state &= ~(SEMAPHORE);
        dequeue_task(mutex);
      }
      rt_schedule();
      return_code = 0;
    }
    break;

  }

  r_restore_flags(flags);
  return(return_code);

}  // End function - pthread_mutex_unlock


// ----------------------------------------------------------------------------

// Conditional Variables section.


int pthread_cond_init(pthread_cond_t *cond,
                      const pthread_condattr_t *cond_attr) {

  QUEUE *rt_q;

  cond->c_waiting.magic = RT_SEM_MAGIC;
  rt_q = &(cond->c_waiting.queue);
  rt_q->prev = &(cond->c_waiting.queue);
  rt_q->next = &(cond->c_waiting.queue);
  rt_q->task = 0;
  return(0);

}  // End function - pthread_cond_init


int pthread_cond_destroy(pthread_cond_t *cond) {

  unsigned long flags;
  int return_code;
  QUEUE *rt_q;

  rt_q = &(cond->c_waiting.queue);
  r_save_flags(flags);
  r_cli();
  if( (rt_q->next != &(cond->c_waiting.queue)) &&
                     (rt_q->prev != &(cond->c_waiting.queue)) ) {
    return_code = EBUSY;
  } else {
    cond->c_waiting.magic = 0;
    return_code = 0;
  }

  r_restore_flags(flags);
  return(return_code);

}  // End function - pthread_cond_destroy



int pthread_condattr_init(pthread_condattr_t *attr) {

  return(0);

}  // End function - pthread_condattr_init



int pthread_condattr_destroy(pthread_condattr_t *attr) {

  return(0);

}  // End function - pthread_condattr_destroy


int pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex) {

  unsigned long flags;
  QUEUE *rt_q;
  RT_TASK *current_task;

  current_task = rt_whoami();
  r_save_flags(flags);
  r_cli();
  current_task->state |= SEMAPHORE;
  current_task->blocked_on.sem = &(cond->c_waiting);
  cond_enqueue_task(cond);
  pthread_mutex_unlock(mutex);
  rt_schedule();
  pthread_mutex_lock(mutex);
  r_restore_flags(flags);
  return(0);

}  // End function - pthread_cond_wait


int pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex,
                           const struct timespec *abstime) {

  unsigned long flags;
  QUEUE *rt_q;
  RT_TASK *current_task;

//  TODO
// Gotta figure out how to get absoulte time in RT land, sigh....
  return(ETIMEDOUT);

}  // End function - pthread_cond_timedwait


int pthread_cond_signal(pthread_cond_t *cond) {

  unsigned long flags;
  QUEUE *rt_q;
  RT_TASK *next_task;

  r_save_flags(flags);
  r_cli();
  if(cond->c_waiting.queue.next != &(cond->c_waiting.queue) &&
     cond->c_waiting.queue.next != NULL && cond->c_waiting.queue.prev != NULL) {

    cond->c_waiting.queue.next->task->blocked_on.sem = 0;
    cond->c_waiting.queue.next->task->state &= ~(SEMAPHORE);
    next_task = (cond->c_waiting.queue.next)->task;
    (next_task->queue.prev)->next = next_task->queue.next;
    (next_task->queue.next)->prev = next_task->queue.prev;

  }

  rt_schedule();
  r_restore_flags(flags);
  return(0);

}  // End function - pthread_cond_signal


int pthread_cond_broadcast(pthread_cond_t *cond) {

  unsigned long flags;
  QUEUE *rt_q;
  RT_TASK *next_task;

  r_save_flags(flags);
  r_cli();
  if(cond->c_waiting.queue.next != NULL && cond->c_waiting.queue.prev != NULL) {

    while(cond->c_waiting.queue.next != &(cond->c_waiting.queue)) {

      cond->c_waiting.queue.next->task->blocked_on.sem = 0;
      cond->c_waiting.queue.next->task->state &= ~(SEMAPHORE);
      next_task = (cond->c_waiting.queue.next)->task;
      (next_task->queue.prev)->next = next_task->queue.next;
      (next_task->queue.next)->prev = next_task->queue.prev;

    }

  }

  rt_schedule();
  r_restore_flags(flags);
  return(0);

}  // End function - pthread_cond_broadcast





// ----------------------------------------------------------------------------


int init_module(void) {

// Although pthreads are not periodic by default, needed by the scheduler?
  tick_period = start_rt_timer((int)nano2count(DEFAULT_TICK_PERIOD));
  rt_set_periodic_mode();
  rt_linux_use_fpu(TASK_FPU_ENABLE);

  printk("\n\n==== RT POSIX Threads API v%s Loaded. ====\n\n", pthread_version);

  return(0);

}  // End function - init_module


void cleanup_module(void) {

  int i;

  stop_rt_timer();
  printk("\n==== RT POSIX Threads API Unloaded. ====\n\n");

}  // End function - cleanup_module


// ---------------------------------< eof >------------------------------------
